from typing import Optional

from genologics.entities import Artifact
from pydantic import BaseModel, Field, validator


class CovidPrepFileRow(BaseModel):
    lims_id: str = Field(alias="LIMS ID")
    sample_well: str = Field(alias="Sample Well")
    destination_well: str = Field(alias="Destination Well")
    index_well: str = Field(alias="Index Well")

    class Config:
        allow_population_by_field_name = True
        # arbitrary_types_allowed = True
        # validate_assignment = True


class BarcodeFileRow(BaseModel):
    source_artifact: Artifact
    destination_artifact: Artifact
    pool: bool
    buffer: bool
    source_labware: Optional[str] = Field(alias="Source Labware")
    barcode_source_container: Optional[str] = Field(alias="Barcode Source Container")
    source_well: Optional[str] = Field(alias="Source Well")
    sample_volume: str = Field(alias="Sample Volume")
    destination_labware: Optional[str] = Field(alias="Destination Labware")
    barcode_destination_container: Optional[str] = Field(alias="Barcode Destination Container")
    destination_well: Optional[str] = Field(alias="Destination Well")
    buffer_volume: str = Field(alias="Buffer Volume")

    @validator("barcode_source_container", always=True, pre=True)
    def set_barcode_source_container(cls, v, values: dict) -> Optional[str]:
        return values["source_artifact"].udf.get("Output Container Barcode")

    @validator("barcode_destination_container", always=True, pre=True)
    def set_barcode_destination_container(cls, v, values: dict) -> Optional[str]:
        return values["destination_artifact"].udf.get("Output Container Barcode")

    @validator("source_labware", always=True, pre=True)
    def set_source_labware(cls, v, values: dict) -> Optional[str]:
        return values["source_artifact"].location[0].type.name

    @validator("destination_labware", always=True, pre=True)
    def set_sdestination_labware(cls, v, values: dict) -> Optional[str]:
        return values["destination_artifact"].location[0].type.name

    @validator("sample_volume", always=True, pre=True)
    def set_sample_volume(cls, v, values: dict) -> str:
        if values["pool"]:
            return values["source_artifact"].udf.get(v)
        else:
            return values["destination_artifact"].udf.get(v)

    @validator("buffer_volume", always=True, pre=True)
    def set_buffer_volume(cls, v, values: dict) -> str:
        if values["buffer"]:
            return values["destination_artifact"].udf.get(v)
        else:
            return 0

    @validator("source_well", always=True, pre=True)
    def set_source_well(cls, v, values: dict) -> str:
        if values["source_labware"] == "Tube":
            return values["barcode_source_container"]
        else:
            return values["source_artifact"].location[1].replace(":", "")

    @validator("destination_well", always=True, pre=True)
    def set_destination_well(cls, v, values: dict) -> str:
        if values["destination_labware"] == "Tube":
            return values["barcode_destination_container"]
        else:
            return values["destination_artifact"].location[1].replace(":", "")

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        validate_assignment = True



import json

import requests


class StatusDBAPI(object):
    def __init__(self, url=None):
        self.url = url

    def apptag(self, tag_name, key=None, entry_point="/applications"):
        res = requests.get(self.url + entry_point + "/" + tag_name)
        if key:
            return json.loads(res.text)[key]
        else:
            return json.loads(res.text)



from pydantic import BaseModel, Field

from cg_lims.models.api.master_steps import HybridizeLibraryTWIST


class DebugKapaCSV(BaseModel):
    sample_id: str = Field(..., alias="SampleID")
    hyb_bait_set: str = Field(None, alias="HybridizeBaitSet")
    hyb_capture_kit: str = Field(None, alias="HybridizeCaptureKit")
    hyb_container_name: str = Field(None, alias="HybridizeContainerName")
    hyb_enrichment_kit: str = Field(None, alias="HybridizeTwistEnrichmentKit")
    hyb_blockers: str = Field(None, alias="HybridizeBlockers")
    hyb_kit: str = Field(None, alias="HybridizeKit")
    hyb_thermal_cycler: str = Field(None, alias="HybridizeThermalCycler")
    hyb_method_document: str = Field(None, alias="HybridizeMethodDocument")
    hyb_document_version: str = Field(None, alias="HybridizeDocumentVersion")
    aliquot_performance_NA24143: str = Field(None, alias="AliquotBatchNoPrepPerformanceNA24143")
    aliquot_GMCKsolid_HD827: str = Field(None, alias="AliquotBatchNoGMCKsolidHD827")
    aliquot_GMSlymphoid_HD829: str = Field(None, alias="AliquotBatchNoGMSlymphoidHD829")
    aliquot_GMSmyeloid_HD829: str = Field(None, alias="AliquotBatchNoGMSmyeloidHD829")
    aliquot_amount_needed: str = Field(None, alias="AliquotAmountNeeded")
    kapa_label: str = Field(None, alias="KapaLabel")
    kapa_method_document: str = Field(None, alias="KapaMethodDocument")
    kapa_document_version: str = Field(None, alias="KapaDocumentVersion")
    kapa_kit: str = Field(None, alias="KapaPrepKit")
    kapa_well: str = Field(None, alias="KapaWell")
    kapa_container_name: str = Field(None, alias="KapaContainerName")
    kapa_size_bp: str = Field(None, alias="PreHybLibrarySize")
    kapa_concentration: str = Field(None, alias="PreHybConcentration")
    pool_amount_of_sample: str = Field(None, alias="PoolAmount")
    capt_enrichment_kit: str = Field(None, alias="CaptureEnrichmentKit")
    capt_hybridization_time: str = Field(None, alias="CaptureHybridizationTime")
    bead_size_bp: str = Field(None, alias="PostHybLibrarySize")
    bead_concentration: str = Field(None, alias="PostHybConcentration")
    buff_concentration: str = Field(None, alias="BufferExchangeConcentration")

    class Config:
        arbitrary_types_allowed = True

    def set_hybridize(self, hybridize: HybridizeLibraryTWIST):
        self.hyb_bait_set = hybridize.bait_set
        self.hyb_capture_kit = hybridize.capture_kit
        self.hyb_container_name = hybridize.container_name
        self.hyb_enrichment_kit = hybridize.enrichment_kit
        self.hyb_blockers = hybridize.blockers
        self.hyb_kit = hybridize.hybridization_kit
        self.hyb_thermal_cycler = hybridize.thermal_cycler
        self.hyb_method_document = hybridize.method_document
        self.hyb_document_version = hybridize.document_version

    def set_aliquot(self, aliquot):
        self.aliquot_performance_NA24143 = aliquot.performance_NA24143
        self.aliquot_GMCKsolid_HD827 = aliquot.GMCKsolid_HD827
        self.aliquot_GMSlymphoid_HD829 = aliquot.GMSlymphoid_HD829
        self.aliquot_GMSmyeloid_HD829 = aliquot.GMSmyeloid_HD829
        self.aliquot_amount_needed = aliquot.amount_needed

    def set_kapa(self, kapa):
        self.kapa_label = kapa.label
        self.kapa_method_document = kapa.method_document
        self.kapa_document_version = kapa.document_version
        self.kapa_kit = kapa.prep_kit
        self.kapa_well = kapa.well
        self.kapa_container_name = kapa.container_name
        self.kapa_size_bp = kapa.size_bp
        self.kapa_concentration = kapa.concentration

    def set_pool(self, pool):
        self.pool_amount_of_sample = pool.amount_of_sample

    def set_capture(self, capture):
        self.capt_enrichment_kit = capture.enrichment_kit
        self.capt_hybridization_time = capture.hybridization_time

    def set_bead(self, bead):
        self.bead_size_bp = bead.size_bp
        self.bead_concentration = bead.concentration

    def set_buffer(self, buffer):
        self.buff_concentration = buffer.concentration


HEADERS = [
    "SampleID",
    "HybridizeBaitSet",
    "HybridizeCaptureKit",
    "HybridizeContainerName",
    "HybridizeTwistEnrichmentKit",
    "HybridizeBlockers",
    "HybridizeKit",
    "HybridizeThermalCycler",
    "HybridizeMethodDocument",
    "HybridizeDocumentVersion",
    "AliquotBatchNoPrepPerformanceNA24143",
    "AliquotBatchNoGMCKsolidHD827",
    "AliquotBatchNoGMSlymphoidHD829",
    "AliquotBatchNoGMSmyeloidHD829",
    "AliquotAmountNeeded",
    "KapaLabel",
    "KapaMethodDocument",
    "KapaDocumentVersion",
    "KapaPrepKit",
    "KapaWell",
    "KapaContainerName",
    "PreHybLibrarySize",
    "PreHybConcentration",
    "PoolAmount",
    "CaptureEnrichmentKit",
    "CaptureHybridizationTime",
    "BufferExchangeConcentration",
    "PostHybLibrarySize",
    "PostHybConcentration",
]



#!/usr/bin/env python
import click
import yaml
from genologics.lims import Lims

from cg_lims import options
from cg_lims.status_db_api import StatusDBAPI

# commands
from cg_lims.EPPs import epps
from cg_lims.commands.serve import serve_command


@click.group(invoke_without_command=True)
@options.config()
@click.pass_context
def cli(ctx, config):
    with open(config) as file:
        config_data = yaml.load(file, Loader=yaml.FullLoader)
    lims = Lims(config_data["BASEURI"], config_data["USERNAME"], config_data["PASSWORD"])
    status_db = StatusDBAPI(config_data["CG_URL"])

    ctx.ensure_object(dict)
    ctx.obj["lims"] = lims
    ctx.obj["status_db"] = status_db
    ctx.obj["host"] = config_data.get("CG_LIMS_HOST")
    ctx.obj["port"] = config_data.get("CG_LIMS_PORT")


cli.add_command(epps)
cli.add_command(serve_command)



#!/usr/bin/env python

import click

# commands
from cg_lims.EPPs.udf.copy.artifact_to_sample import artifact_to_sample
from cg_lims.EPPs.udf.copy.sample_to_artifact import sample_to_artifact
from cg_lims.EPPs.udf.copy.process_to_sample import process_to_sample


@click.group(invoke_without_command=True)
@click.pass_context
def copy(ctx):
    """Main entry point of copy commands"""
    pass


copy.add_command(artifact_to_sample)
copy.add_command(sample_to_artifact)
copy.add_command(process_to_sample)



import logging
import sys
from typing import List

import click
from genologics.entities import Artifact, Process

from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)


def calculate_volumes(artifacts: List[Artifact], process: Process):
    """Calculates volumes for diluting samples. The total volume differ depending on
    type of sample. It is given by the process udf 'Total Volume (ul)'."""

    max_volume = process.udf.get("Total Volume (ul)")
    if not max_volume:
        raise MissingUDFsError("Process udf missing: Total Volume (ul)")

    missing_udfs = 0

    for art in artifacts:
        concentration = art.udf.get("Concentration")
        amount = art.udf.get("Amount needed (ng)")
        if not amount or not concentration:
            missing_udfs += 1
            continue

        art.udf["Sample Volume (ul)"] = amount / float(concentration)
        art.udf["Volume H2O (ul)"] = max_volume - art.udf["Sample Volume (ul)"]
        art.put()

    if missing_udfs:
        raise MissingUDFsError(f"Udf missing for {missing_udfs} samples")


@click.command()
@click.pass_context
def twist_aliquot_volume(ctx):
    """Calculates amount needed for samples."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    process = ctx.obj["process"]

    try:
        artifacts = get_artifacts(process=process, input=False)
        calculate_volumes(artifacts=artifacts, process=process)
        message = "Volumes have been calculated for all samples."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



#!/usr/bin/env python
import click

from cg_lims.EPPs.qc.set_qc_fail import set_qc_fail


@click.group(invoke_without_command=True)
@click.pass_context
def qc(ctx):
    """Main entry point of move commands"""
    pass


qc.add_command(set_qc_fail)



import pathlib
from pathlib import Path
from typing import Optional

from genologics.config import BASEURI
from genologics.entities import Artifact, Process
from genologics.lims import Lims

from cg_lims.exceptions import FileError


def get_log_content(lims: Lims, file_id: str) -> str:
    """Searching for a log file Artifact with file_id.

    If the artifact is found, returning its file content."""

    log_artifact = Artifact(lims, id=file_id)
    file_path = get_file_path(log_artifact)

    if file_path:
        log_file = pathlib.Path(file_path)
        log_file_content = log_file.read_text()
    else:
        log_file_content = ""

    return log_file_content


def get_file_path(file_artifact: Artifact) -> Optional[str]:
    """getting the file path from a artifacts content location"""

    try:
        files = file_artifact.files
    except:
        files = None

    if files:
        server_adress = BASEURI.split(":")[1]
        file_path = files[0].content_location.split(server_adress)[1]
        return file_path
    else:
        return None



from genologics.lims import Lims
from typing import Optional
from datetime import date


def get_udf_type(lims: Lims, udf_name: str, attach_to_name: str) -> Optional:
    """Get udf type.

    Args:
        udf_name:
            eg: 'Concentration (nM)', 'Comment',...
        attach_to_name: name of entity to which the udf is attached
            eg. 'ResultFile', 'Analyte', 'Aggregate QC (RNA)',...
    """
    udf_types = {"String": str, "Numeric": float, "Date": date}

    udf_configs = lims.get_udfs(name=udf_name, attach_to_name=attach_to_name)
    udf_configs[0].get()
    udf_type = udf_configs[0].root.attrib["type"]

    return udf_types[udf_type]



"""Base cg_lims calss definitions
"""

from genologics.entities import Artifact

from cg_lims.exceptions import MissingUDFsError, ZeroReadsError


class Pool:
    """Collecting informatiuon general to any pool"""

    def __init__(self, pool_artifact: Artifact):
        self.total_reads_missing = 0
        self.artifact = pool_artifact

    def get_total_reads_missing(self) -> None:
        """Get the total numer of missing reads in the pool"""

        for art in self.artifact.input_artifact_list():
            reads = art.samples[0].udf.get("Reads missing (M)")
            if reads is None:
                raise MissingUDFsError("Missing udfs: Reads missing (M)")
            self.total_reads_missing += reads
        if self.total_reads_missing == 0:
            raise ZeroReadsError("All samples seem to have Missing Reads = 0!")



import logging
import sys
from typing import List

import click

from cg_lims import options
from cg_lims.get.artifacts import get_latest_artifact, get_qc_output_artifacts

LOG = logging.getLogger(__name__)


@click.command()
@options.process_type(help="Get buffer from this process type(s)")
@click.pass_context
def get_volumes_from_buffer(ctx, process_type: List[str]) -> None:
    """Getting Volume Elution from previous step of type defined by process_types.
    If volume found, setting the value Volume udf on artifact of current step. As part of the sample is used in the QC,
    the value is subtracted by 10 to get the actual volume that is left."""

    LOG.info(f"Running {ctx.command_path} with params {ctx.params}")

    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    failed_count = 0
    updated_count = 0
    artifacts = get_qc_output_artifacts(lims, process)
    for artifact in artifacts:
        try:
            buffer_artifact = get_latest_artifact(
                lims=lims, sample_id=artifact.samples[0].id, process_type=list(process_type)
            )
        except:
            continue
        if buffer_artifact.udf.get("Volume Elution (ul)"):
            volume_buffer = float(buffer_artifact.udf.get("Volume Elution (ul)")) - 10.0
            if volume_buffer < 0:
                failed_count += 1
                continue
            updated_count += 1
            artifact.udf["Volume (ul)"] = volume_buffer
            artifact.put()

    message = f"Updated {updated_count} samples with volume from Buffer step."
    LOG.info(message)
    click.echo(message)

    if failed_count > 0:
        message = f"Failed to update {failed_count} artifacts - buffer volume lower than 0!"
        LOG.error(message)
        sys.exit(message)



import click
import logging
import uvicorn

from cg_lims.models.context import APIContextObject

LOG = logging.getLogger(__name__)


@click.command("serve")
@click.option("--reload", is_flag=True)
@click.pass_context
def serve_command(ctx, reload: bool):
    """Serve the Statina app for testing purpose.
    This command will serve the user interface (external) as default
    """
    api_context = APIContextObject(**ctx.obj)
    LOG.info("Running api on host:%s and port:%s", api_context.host, api_context.port)
    uvicorn.run(
        app="cg_lims.app.api.api_v1.api:app",
        host=api_context.host,
        port=api_context.port,
        reload=reload,
    )



from typing import Optional

from genologics.entities import Artifact
from pydantic import Field, validator

from cg_lims.models.api.master_steps.base_step import get_artifact_udf, BaseStep
from cg_lims.get.artifacts import get_latest_artifact


class BufferExchange(BaseStep):
    artifact: Optional[Artifact]
    concentration: Optional[str] = Field(None, alias="Concentration")

    @validator("artifact", always=True)
    def get_artifact(cls, v, values):
        try:
            return get_latest_artifact(
                lims=values.get("lims"),
                sample_id=values.get("sample_id"),
                process_type=["Buffer Exchange v2"],
            )
        except:
            return None

    @validator("concentration", always=True)
    def get_size_bp(cls, v, values):
        return get_artifact_udf(values.get("artifact"), "Concentration")



from operator import attrgetter
from typing import List

from genologics.entities import Artifact, Process, Sample
from genologics.lims import Lims

from cg_lims.exceptions import FileError, MissingArtifactError


def get_sample_artifact(lims: Lims, sample: Sample) -> Artifact:
    """Returning the initial artifact related to a sample.
    Assuming first artifact is allways named sample.id + 'PA1."""

    return Artifact(lims, id=f"{sample.id}PA1")


def get_artifacts(process: Process, input: bool) -> List[Artifact]:
    """If inputs is True, returning all input analytes of the process,
    otherwise returning all output analytes of the process"""

    if input:
        artifacts = process.all_inputs(unique=True)
    else:
        artifacts = [a for a in process.all_outputs(unique=True) if a.type == "Analyte"]
    return artifacts


def get_artifact_by_name(process: Process, name: str) -> Artifact:
    """Searches for output artifacts with name <file_name>
    Raises error if more than one artifact with same name!"""

    artifacts = list(filter(lambda a: a.name in [name], process.all_outputs()))

    if len(artifacts) > 1:
        raise FileError(f"more than one output artifact named {name}")

    return artifacts[0]


def get_qc_output_artifacts(lims: Lims, process: Process) -> List[Artifact]:
    """Get output 'artifacts' (messuements) of a qc process"""

    input_output_maps = process.input_output_maps
    artifact_ids = [
        io[1]["limsid"] for io in input_output_maps if io[1]["output-generation-type"] == "PerInput"
    ]
    return [Artifact(lims, id=id) for id in artifact_ids if id is not None]


def filter_artifacts(artifacts: List[Artifact], udf: str, value) -> List[Artifact]:
    """Returning a list of only artifacts with udf==value"""

    return [a for a in artifacts if a.udf.get(udf) == value]


def get_latest_artifact(lims: Lims, sample_id: str, process_type: List[str]) -> Artifact:
    """Getting the most recently generated artifact by process_type and sample_id.

    Searching for all artifacts (Analytes) associated with <sample_id> that
    were produced by <process_type>.

    Returning the artifact with latest parent_process.date_run.
    If there are many such artifacts only one will be returned."""

    artifacts = lims.get_artifacts(
        samplelimsid=sample_id,
        type="Analyte",
        process_type=process_type,
    )
    if not artifacts:
        raise MissingArtifactError(
            f"Could not find a artifact with sample: {sample_id} generated by process: {' ,'.join(process_type)}"
        )

    artifacts = [(a.parent_process.date_run, a.id, a) for a in artifacts]
    artifacts.sort()
    date, id, latest_art = artifacts[-1]

    return latest_art



from typing import List

from genologics.entities import Artifact, Process, Sample

from cg_lims.exceptions import MissingSampleError


def get_process_samples(process: Process) -> List[Sample]:
    """Get all samples in a process"""

    all_samples = []
    for art in process.all_inputs():
        all_samples += art.samples

    return list(set(all_samples))


def get_one_sample_from_artifact(artifact: Artifact) -> Sample:
    """Checking that an artifact has one and only one sample.
    Returning the sample if it exists.
    Raising MissingSampleError otherwise."""

    try:
        samples = artifact.samples
    except:
        raise MissingSampleError(f"Artifact {artifact.id} has no samples.")
    if not samples:
        raise MissingSampleError(f"Artifact {artifact.id} has no samples.")
    elif len(samples) > 1:
        raise MissingSampleError(f"Artifact {artifact.id} has more than one sample.")

    return samples[0]



from fastapi import APIRouter, Depends
from cg_lims.app.api.deps import get_lims
from cg_lims.models.api import Sample

from genologics.lims import Lims
from genologics.entities import Sample as GenologicsSample


router = APIRouter()


@router.get("/{sample_id}", response_model=Sample)
def get_sample(sample_id: str, lims: Lims = Depends(get_lims)):
    """get sample"""
    sample = GenologicsSample(lims, id=sample_id)
    sample_udfs = dict(sample.udf.items())
    return Sample(**sample_udfs, id=sample.id, name=sample.name, project=sample.project.id)
